<?php
if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

/**
 * WorthstockEnqueue è una classe che raccoglie in coda le azioni che invoca una chiamata API WS remota
 * 
 * Questa classe riceve gli eventi di insert, cancellazione e update che riguardano i prodotti WC per cui deve essere effettuata attraverso le API di WS una chiamata remota CRUD.
 * Gli eventi sono normalmente collegati agli hook di WC. Per esempio un'azione di update in Woocommerce viene rilevata attraverso un hook che invoca un add_update_event() in questa classe. Questo tipo di funzione effettua dei controlli sulla richiesta per capire se va fatta passare e in caso positivo aggiunge l'azione a una coda locale che poi viene risolta prima che Wordpress/PHP concluda la sua richiesta e restituisca un risultato.
 */
class WorthstockEnqueue extends WorthstockBase {    
    /**
     * @var array
     */
    private $event_queue;
    private $remote_products;
    const API_CALL_PER_SECOND = 10;
    const API_SEC_DELAY = 0.2;

    public function __construct()
    {
        $this->event_queue = [];
    }

    public function add_event_to_queue($event_type, $product_id, $origin="bulk", $payload=null) {
        // check if already enqueued
        // if ($this->is_event_enqueued($product_id, $event_type) === true) return;
        // filter out products in draft and other cases from sync //TODO: forse va spostato altrove
        if ($this->is_product_to_be_synched($product_id, $event_type, $origin) !== true) return;
        $this->event_queue[$product_id][] = $event_type;
    }

    public function add_delete_event($product_id) {
        $this->add_event_to_queue('delete', $product_id);
    }

    public function add_insert_event($product_id) {
        $this->add_event_to_queue('insert', $product_id);
    }

    public function add_update_event($product_id) {
        $this->add_event_to_queue('update', $product_id);
    }

    public function add_price_event($product_id) {
        $this->add_event_to_queue('price', $product_id);
    }

    public function add_stock_event($product_id) {
        if (is_object($product_id)) {
            $product_id = $product_id->get_id();
        }
        $this->add_event_to_queue('stock', $product_id);
    }

    public function get_event_queue() {
        return $this->event_queue;
    }

    public function is_event_enqueued($product_id, $event_type) {
        $queue = $this->get_event_queue();
        return ( isset($queue[$product_id]) && in_array($event_type, $queue[$product_id]) ) ? true : false;
    }

    /**
     * il prodotto rispetta i requisiti per richiedere un aggiornamento?
     */
    public function is_product_to_be_synched($product_id, $event_type=null, $origin) {
	    $post = get_post( $product_id );
        // if setting automatic replication is off and ain't bulk operation and the product ain't already synced, skip      
        if ($origin != 'bulk' && $this->main()->settings->get_option( 'worthstock_automatic_replication' ) == 'no' && $this->main()->mapper->isProductSynced($post) === false) {
            return;            
        }
        // Not drafts
        if ( ! in_array($post->post_status, array('publish', 'private')) ) return;
        // Only products
        if ( $post->post_type != 'product') return;
         // Exit if it's an autosave
        if ( defined( 'DOING_AUTOSAVE' ) && DOING_AUTOSAVE ) return;
        // check if remote products still existing
        $this->check_product_expired_ids($product_id);
        // return if product has not changed
        if ( ($event_type == 'insert' || $event_type == 'update') && $this->has_product_changed($product_id, $event_type) !== true) return;

        return true;        
    }
    
    public function has_product_changed($product_id, $event_type) {
        $metakey = '_wst_temp_serialized_product';
        if ( !($temp_serialized_product = get_post_meta($product_id, $metakey, true)) ) return true;
        $product = wc_get_product( $product_id );
        $action_type = ($this->main()->mapper->isProductSynced($product)) ? 'update' : 'insert';
        $serialized_product = $this->main()->mapper->convertProductToWorthstock($product_id, $action_type);
        $diff = $this->array_diff_assoc_recursive($serialized_product, $temp_serialized_product);
        $diff = $this->array_extract_keys($diff);
        foreach ($diff as $idx=>$value) {
            if ( is_int($value) ) {
                unset($diff[$idx]);
            } else
            if ( $value == 'IdProductVariant' ) {
                unset($diff[$idx]);
            } else
            if ( $value == 'Variants' ) {
                unset($diff[$idx]);
            } else
            if ( stristr($value, 'quantity') ) {
                $this->add_stock_event($product_id);
                unset($diff[$idx]);
            } else
            if ( stristr($value, 'price') ) {
                $this->add_price_event($product_id);
                unset($diff[$idx]);
            }
        }
        if (count(array_filter($diff)) == 0) return false;

        //$this->clear_serialized_product($product_id);

        return true;
    }

    public function save_serialized_product($post_id, $post_data=null) {
        if ( 'product' !== get_post_type($post_id) ) {
			return;
		}
        $product = wc_get_product( $post_id );
        $metakey = '_wst_temp_serialized_product';
        $action_type = ($this->main()->mapper->isProductSynced($product)) ? 'update' : 'insert';
        @$temp_serialized_product = $this->main()->mapper->convertProductToWorthstock($post_id, $action_type);
        update_post_meta($post_id, $metakey, $temp_serialized_product);
    }

    private function clear_serialized_product($post_id) {
        $metakey = '_wst_temp_serialized_product';
        delete_post_meta($post_id, $metakey);
    }

    /**
     * prima di ogni altra operazione controllo che il prodotto abbia una corrispondenza su WS, facendo una chiamata di controllo sugli IdProduct
     * Quelli inesistenti lato remoto vengono rimossi dai custom meta
     */
    private function check_product_expired_ids($product_id) {
        // suppose there may be multiples ids for product
        $id_worthstock = get_post_meta($product_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));
        if (count(array_filter($id_worthstock)) > 0) {
        //if (!empty($id_worthstock)) {
            $id_worthstock = (is_array($id_worthstock)) ? $id_worthstock : [$id_worthstock];
            // get remote existing product with said ids
            // credo una cache locale degli id_products verificati, mancando una cache di basso livello delle chiamate API
            if ( isset($this->remote_products[$product_id]) ) {
                $id_products = $this->remote_products[$product_id];
                $diffed = (array_diff($id_worthstock, $id_products));
            } else {
                $existing_products = $this->main()->sync->api_get_all_products_drafted($id_worthstock);
                if ($existing_products && is_array($existing_products)) {
                    $id_products = wc_list_pluck($existing_products['Products'], 'IdProduct');                    
                    $this->remote_products[$product_id] = $id_products;
                    $diffed = (array_diff($id_worthstock, $id_products));
                }
            }
        }

        if (count(array_filter($id_worthstock)) == 0 || count($diffed) ) {
            delete_post_meta($product_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));
            $this->clear_serialized_product($product_id);
        }
        
        if ( count($diffed) && is_array($id_products) && count($id_products) > 0 ) {   
            delete_post_meta($product_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'));      
            foreach ($id_products as $id) {
                // add post meta if present in original array (preventing possible wrong conditions)
                if (in_array($id, $id_worthstock)) {
                    add_post_meta( $product_id, $this->main()->mapper->getWorthstockCustomParamName('id_product'), $id);
                }
            }
        }

    }

    public function array_diff_assoc_recursive($array1, $array2) {
        $difference = array();
        //$skip_keys = array("QuantityAvailable","IdProductVariant","PriceLST","PriceB2C","PricePROMO","PriceOVER","UniqueIdShopify","IdProduct");
        foreach($array1 as $key => $value) {
            //if (in_array($key, $skip_keys, true)) continue;
            if( is_array($value) ) {
                if( !isset($array2[$key]) || !is_array($array2[$key]) ) {
                    $difference[$key] = $value;
                } else {
                    $new_diff = $this->array_diff_assoc_recursive($value, $array2[$key]);
                    if( !empty($new_diff) )
                        $difference[$key] = $new_diff;
                }
            } else if( !array_key_exists($key,$array2) || $array2[$key] !== $value ) {
                $difference[$key] = $value;
            }
        }
        return $difference;
    }

    function array_flatten(array $array) {
        $return = array();
        array_walk_recursive($array, function($a) use (&$return) { $return[] = $a; });
        return $return;
    }

    function array_extract_keys($ar) {           
        foreach($ar as $k => $v) {
            //if ($k) {
                $keys[] = $k;
            //}
            if (is_array($ar[$k])) {
                $keys = array_merge($keys, $this->array_extract_keys($ar[$k]));
            }
        }
        return $keys;
    }

    /**
     * elaboro la coda per fare i controlli preliminari
     */
    /* private function queue_prechecks() {        
        if (! $queue = $this->get_event_queue() ) return false;
    } */

    /**
     * srotolo le azioni messe in coda e per ogni prodotto filtro solo l'ultima (gli hook di woocommerce possono essere richiamati più volte in modo ridondante)
     */
    public function resolve_queue() {
        if (! $queue = $this->get_event_queue() ) return false;
        if ($this->main()->get_woocommerce_replication() !== true) return false;
        $this->main()->sync->api_logger( $this->get_event_queue() );
        if (is_array($queue) && count($queue)) {
            // $product_ids = array_keys($queue);
            // $product_id = end($product_ids);
            //$last_action_in_queue = end(end($queue));
            foreach($queue as $product_id=>$actions) {
                // filtro le action per prodotto in modo da eliminare doppioni
                $actions = array_unique($actions);
                //$last_action_in_queue = end($actions);
                $this->main()->sync->api_logger( 'Product_id:' . $product_id . ' Action:' . $actions);
                foreach($actions as $action) {
                    if ($product_id && $action) {
                        //TODO: setting provvisorio per disabilitare aggiornamento automatico
                        if ($this->main()->settings->get_option( 'worthstock_enable_replication' ) == 'yes') {
                            // ferma esecuzione per questa frazione di secondo tra una chiamata e l'altra
                            if (self::API_SEC_DELAY) {
                                usleep( (self::API_SEC_DELAY * 1000000) );
                            }
                            $this->exec_notification($product_id, $action);
                        }
                    }
                }
            }
        }
    }

    /**
     * this invoke the function performing actual api notification
     */
    protected function exec_notification($product_id, $notification_type) {
        switch( $notification_type ) {
            case 'insert':
                $this->main()->sync->notify_product_inserted($product_id);
                break;

            case 'delete':
                $this->main()->sync->notify_product_deleted($product_id);
                break;

            case 'update':
                $this->main()->sync->notify_product_updated($product_id);
                break;

            case 'price':
                $this->main()->sync->notify_price_updated($product_id);
                break;

            case 'stock':
                $this->main()->sync->notify_stock_updated($product_id);
                break;

            default:
                return null;
        }

    }
}